# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-09-24 11:29
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ACPHoliday',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateField(unique=True, verbose_name='Date')),
            ],
            options={
                'ordering': ('date',),
            },
        ),
        migrations.CreateModel(
            name='HolidayType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=60, verbose_name='Holiday name')),
                ('short_name', models.CharField(max_length=60, unique=True, verbose_name='short name')),
            ],
        ),
        migrations.AddField(
            model_name='acpholiday',
            name='holiday_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='acp_calendar.HolidayType', verbose_name='Holiday type'),
        ),
    ]




# -*- coding: utf-8 -*-
import collections
import json
import datetime
from django.core.management import BaseCommand

from ... import app_settings
from ...initial_data import get_holidays_dictionary, get_holiday_type_list, \
    get_holidays_list
from ...exceptions import ACPCalendarException
from ...models import ACPHoliday, HolidayType


class Command(BaseCommand):
    """
    To list the holidays:

    .. code-block:: bash

        $ python manage.py acp_holidays --list-initial

    You will get a print out like this. The [*] on the first column means that the initial data
    is already on the database.  The [-] on the first column means that the initial data
    is not on the database.

    .. code-block:: bash

        Year 2006  (11 holidays)
        ----------------------------------------------------------------------
        [*] año_nuevo                      2006-01-01
        [-] mártires                       2006-01-09
        [*] martes_carnaval                2006-02-28
        [*] viernes_santo                  2006-04-14
        [*] día_del_trabajo                2006-05-01

        ...

        ======================================================================
        Found 133 in initials
        Found 132 in database

    To export your the holidays currently in the database to a json file

    .. code-block:: bash

        $  python manage.py acp_holidays --export-holidays=/path/to/your/file.json
    """

    def add_arguments(self, parser):
        #parser.add_argument('optional-argument', nargs='?')
        parser.add_argument('--list-initial',
                            action='store_true',
                            dest='list_initial',
                            default=None,
                            help='List initial data')

        parser.add_argument('--export-holidays',
                            action='store',
                            dest='export_filename',
                            default=None,
                            help='Export holidays in database to json format')
        # parser.add_argument('--variable',
        #                     action='store',
        #                     dest='variable_name',
        #                     default=None,
        #                     help='Useful info')
        # parser.add_argument('--appended-argument',
        #                     action='append',
        #                     dest='appended_arg',
        #                     default=None,
        #                     help='Useful info')
        parser.add_argument('--update-initial',
                            action='store_true',
                            dest='update_initial',
                            default=None,
                            help='Update initial data')
        parser.add_argument('--update-initial-test',
                            action='store_true',
                            dest='update_initial_test',
                            default=None,
                            help='Update initial data (test only)')

    def handle(self, *args, **options):
        if options['list_initial']:
            self._list_initial_data()
        if options['export_filename']:
            self._export_holidays(options['export_filename'])
        if options['update_initial']:
            self._update_initial_data()
        if options['update_initial_test']:
            self._update_initial_data(test=True)

    def _export_holidays(self, filename):
        """
        Exports all ACPHoldays in the database to pretty pring json file in UTF-8 format.
        it exports the holidays in the format the initial_data need to load holidays:

            {
                "date": "2006-01-01",
                "holiday_type": "año_nuevo"
            },
            {
                "date": "2006-01-09",
                "holiday_type": "mártires"
            },
        ..

        :param filename: JSON file to save database content
        """
        holidays = list()
        db_holidays = ACPHoliday.objects.all()
        for db_holiday in db_holidays:
            holiday_dict = dict()
            holiday_dict['date'] = db_holiday.date.strftime(app_settings.LOAD_DATE_FORMAT)
            holiday_dict['holiday_type'] = db_holiday.holiday_type.short_name
            holidays.append(holiday_dict)

        with open(filename, 'w', encoding='utf-8') as outfile:
            json.dump(holidays, outfile, indent=4, ensure_ascii=False)
        self.stdout.write('Wrote %d holidays to %s' % (len(db_holidays), filename))

    def _list_initial_data(self):
        count_initial_holidays = 0
        count_db_holidays = 0

        ordered_holidays = get_holidays_dictionary()
        for year, holidays in ordered_holidays.items():
            self.stdout.write('Year %s  (%d holidays)' % (year, len(holidays)))
            self.stdout.write('-' * 70)
            for holiday in holidays:
                display = dict()
                display['found'] = '*'
                display['date'] = holiday['date']
                count_initial_holidays += 1
                try:
                    assert isinstance(holiday['holiday_type'], str), 'Holiday type should be a string'
                    short_name = holiday['holiday_type']
                    display['holiday_type'] = short_name
                    ACPHoliday.objects.get(holiday_type__short_name=short_name,
                                           date=holiday['date'])
                    count_db_holidays += 1
                except ACPHoliday.DoesNotExist:
                    display['found'] = '-'
                self.stdout.write('\t[{found}] {holiday_type:<30} {date}'.format(**display))
            self.stdout.write('=' * 70)
        self.stdout.write('Found %d in initials' % count_initial_holidays)
        self.stdout.write('Found %d in database' % count_db_holidays)

    def _update_initial_data(self, test=False):
        """Checks for updated holiday types and holidays from initial data and
        adds to database.

        @param: test: Set to True for tests. Will insert some test values and
                      will not write to actual database.
        @type: test: bool
        """
        if test:
            self.stdout.write("WARNING: Test flag is set, will use test data and not write to database.")
        self.stdout.write('Checking for updated HolidayType...')
        initial_holiday_types = get_holiday_type_list()
        if test:
            # Insert a test holiday type
            initial_holiday_types.append(
                {'name': 'TestHolidayType', 'short_name': 'test_holiday_type'})
        self.stdout.write(
            '{} HolidayType found.'.format(len(initial_holiday_types)))
        num_created = 0
        num_updated = 0
        num_skipped = 0
        for ht in initial_holiday_types:
            ht_name = ht.get('name', None)
            ht_short_name = ht.get('short_name', None)
            if not ht_short_name or not ht_name:
                num_skipped += 1
                self.stdout.write('Name or short_name missing: {}'.format(ht))
                continue

            qs = HolidayType.objects.filter(short_name=ht_short_name)
            if not qs.count():
                if not test:
                    HolidayType.objects.create(**ht)
                num_created += 1
                self.stdout.write(
                    'HolidayType {} created.'.format(ht_short_name))
            else:
                # Check if need to update name
                ht_obj = qs.first()
                if ht_obj.name != ht_name:
                    if not test:
                        ht_obj.name = ht_name
                        ht_obj.save()
                    num_updated += 1
                    self.stdout.write(
                        'HolidayType {} name updated to {}.'.format(
                            ht_short_name, ht_name))
        self.stdout.write(
            'HolidayType: {} created, {} updated, {} skipped.'.format(
                num_created, num_updated, num_skipped))

        # Check Holidays
        self.stdout.write('Checking for updated ACPHoliday...')
        holiday_list = get_holidays_list()
        if test:
            # Insert some test holidays
            holiday_list.append({
                "date": "2018-11-01",
                "holiday_type": "test_holiday_type"
            })
            holiday_list.append({
                "date": "2018-12-01",
                "holiday_type": "test_holiday_type"
            })
        self.stdout.write('{} holidays found.'.format(len(holiday_list)))
        num_created = 0
        num_updated = 0
        num_skipped = 0
        for h in holiday_list:
            h_date = h.get('date', None)
            h_short_name = h.get('holiday_type', None)
            # Check both fields have values
            if not h_date or not h_short_name:
                num_skipped += 1
                self.stdout.write('Missing date or holiday_type: {}'.format(h))
                continue
            # Parse date
            try:
                formatted_date = datetime.datetime.strptime(
                    h_date, app_settings.LOAD_DATE_FORMAT)
            except Exception as e:
                num_skipped += 1
                self.stdout.write(
                    'Error parsing date from {}. Exception: {}'.format(h, e))
                continue
            # Check holiday type
            try:
                h_type_obj = HolidayType.objects.get(short_name=h_short_name)
            except HolidayType.DoesNotExist:
                if not test:
                    num_skipped += 1
                    self.stdout.write(
                        'HolidayType {} does not exist. Data {}'.format(
                            h_short_name, h))
                    continue
                else:
                    pass
            # Check holidays
            qs = ACPHoliday.objects.filter(date=formatted_date)
            if qs.count():
                # If holiday exists, check if need to update HolidayType
                h_obj = qs.first()
                if h_obj.holiday_type.short_name != h_short_name:
                    if not test:
                        h_obj.holiday_type = h_type_obj
                        h_obj.save()
                    num_updated += 1
                    self.stdout.write(
                        'ACPHoliday {} updated to HolidayType {}'.format(
                            h_date, h_short_name))
            else:
                # Create a new Holiday
                if not test:
                    h_obj = ACPHoliday.objects.create(
                        date=formatted_date, holiday_type=h_type_obj)
                num_created += 1
                self.stdout.write(
                    'New ACPHoliday created: date: {} holiday_type: {}'.format(
                        h_date, h_short_name))
        self.stdout.write(
            'ACPHoliday: {} created, {} updated, {} skipped.'.format(
                num_created, num_updated, num_skipped))



class AppSettings(object):

    def __init__(self, prefix):
        self.prefix = prefix

    def _settings(self, settings_name, default_value):
        from django.conf import settings
        return getattr(settings, self.prefix + settings_name, default_value)

    @property
    def DATE_FORMAT(self):
        return self._settings('DATE_FORMAT', '%Y-%m-%d')

    @property
    def LOAD_DATE_FORMAT(self):
        """
        This is the date forma for the initial data load. I you change this the date format in
        holiday_initial_data.json must be changed

        :return: String with date format
        """
        return '%Y-%m-%d'

# Ugly? Guido recommends this himself ...
# http://mail.python.org/pipermail/python-ideas/2012-May/014969.html
import sys  # noqa
app_settings = AppSettings('ACP_CALENDAR_')
app_settings.__name__ = __name__
sys.modules[__name__] = app_settings



from django import forms
from django.utils.translation import ugettext_lazy as _


class CalculatorForm(forms.Form):
    start_date = forms.DateField(label=_('Start date'))
    end_date = forms.DateField(label=_('End date'))



# -*- coding: utf-8 -*-
from calendar import monthrange, IllegalMonthError
from datetime import timedelta, date, datetime
from django.utils.translation import ugettext_lazy as _
from django.db import models

from . import app_settings
from .exceptions import ACPCalendarException


class FiscalYear(object):
    """
    This class represents a Pancama Canal Fiscal year which start on October first
    of the previous year and ends on September 30th of the current year.
    For example fiscal year 2016 starts on October 1st, 2015 and ends September 30th
    2016
    """

    def __init__(self, year, **kwargs):
        """
        Contructor for Fiscal year.

        :param year: Fiscal year
        :param kwargs: includes display wich is the template for the str method and
        length which is also used in the str method
        """
        self.year = year
        self.start_date = date(year-1, 10, 1)
        self.end_date = date(year, 9, 30)
        self.fy_format = {'display': kwargs.get('display', 'FY%s'),
                          'length': kwargs.get('length', 2)}

    def __str__(self):
        return self.fy_format['display'] % str(self.year)[self.fy_format['length']:]

    def months_in_fiscal_year(self):
        """
        Gets a tuple of tuple containing the month number and the year of the months in the
        fiscal year.

        :return: a tuple containing 12 tuples. Each tuple contains 2 integer, the first one is the month the
        second one is the year.
        """
        return ((10, self.year-1), (11, self.year-1), (12, self.year-1),
                (1, self.year),(2, self.year),(3, self.year),(4, self.year),
                (5, self.year),(6, self.year),(7, self.year),(8, self.year),
                (9, self.year))

    @staticmethod
    def create_from_date(cdate, **kwargs):
        """
        Creates a Fiscal year object for a date.

        :param cdate: Date or datetime object whitin the fiscal year
        :param kwargs: Same kwargs as for the contructor
        :return: a FiscalYear object
        """
        if isinstance(cdate, datetime):
            cdate = cdate.date()
        start_of_fy = date(cdate.year, 10, 1)
        if cdate >= start_of_fy:
            year = cdate.year + 1
        else:
            year = cdate.year
        return FiscalYear(year, **kwargs)


class HolidayType(models.Model):
    """
    Model for Holiday type.

    .. code-block::python

        holiday = Holiday.objects.create(name='Christmas', short_name='xmas')
    """
    name = models.CharField(_('Holiday name'), max_length=60)
    short_name = models.CharField(_('short name'), max_length=60, unique=True)

    def __str__(self):
        return self.name


class ACPHoliday(models.Model):
    """
    Model for a non working day in the Panama Canal due to a holiday. This model contains all the logic for the
    working days calculations.
    """
    date = models.DateField(_('Date'), unique=True)
    holiday_type = models.ForeignKey(HolidayType, verbose_name=_('Holiday type'))

    def __str__(self):
        return '{0} {1}'.format(self.date.strftime(app_settings.DATE_FORMAT), self.holiday_type)

    class Meta:
        ordering = ('date',)

    @staticmethod
    def validate_dates(start_date, end_date):
        """
        Validates three rules:
        1. End date is not before start date
        2. End date cannot occur after oldest holiday in database
        3. Start date cannot occur before the first holiday in database

        Will raise an ACPCalendarException if one of these rules is broken.

        :param start_date: Start date
        :param end_date: End date
        """
        if start_date > end_date:
            raise ACPCalendarException(_('Start date cannot occur after end date'))
        last_holiday = ACPHoliday.objects.all().last()
        if end_date > last_holiday.date:
            raise ACPCalendarException(_('End date exceed the last registered holiday'))
        first_holiday = ACPHoliday.objects.all().first()
        if start_date < first_holiday.date:
            raise ACPCalendarException(_('Start date precedes the first registered holiday'))

    @staticmethod
    def get_working_days(start_date, end_date, **kwargs):
        """
        Calculates the amount of working days between start date and end date. It will calculate all days that are not
        saturday or sunday and then substract the holiday in the range if they exist.

        :param start_date:
        :param end_date:
        :param kwargs:
        :return: Number of working days between the star date and the end date
        """
        start_date = ACPHoliday.convert_to_date(start_date)
        end_date = ACPHoliday.convert_to_date(end_date)
        ACPHoliday.validate_dates(start_date, end_date)
        day_generator = ACPHoliday.days_in_range_generator(start_date, end_date)
        holidays_in_range = ACPHoliday.objects.filter(date__gte=start_date, date__lte=end_date).count()
        working_days = sum(1 for day in day_generator if day.weekday() < 5)
        return working_days - holidays_in_range

    @staticmethod
    def convert_to_date(study_date):
        """
        Converts String to date to a date object.

        :param study_date: date to be processed
        :return: date object
        """
        if isinstance(study_date, str):
            try:
                date_object = datetime.strptime(study_date, app_settings.DATE_FORMAT).date()
                return date_object
            except ValueError as e:
                raise ACPCalendarException(str(e))
        elif isinstance(study_date, date):
            return study_date
        else:
            raise ACPCalendarException('Dates must be either string or date objects')

    @staticmethod
    def days_in_range_generator(start_date, end_date):
        """
        Creates a generator that contains all date dates between start_date and end_date

        :param start_date: Start date in string or date
        :param end_date: End date in string or date
        :return: A generator containing all dates in range
        """
        start_date = ACPHoliday.convert_to_date(start_date)
        end_date = ACPHoliday.convert_to_date(end_date)
        start_date = start_date - timedelta(1)
        day_generator = (start_date + timedelta(x + 1) for x in range((end_date - start_date).days))
        return day_generator

    @staticmethod
    def week_end_days(start_date, end_date):
        start_date = ACPHoliday.convert_to_date(start_date)
        end_date = ACPHoliday.convert_to_date(end_date)
        day_generator = ACPHoliday.days_in_range_generator(start_date, end_date)
        week_end_days = sum(1 for day in day_generator if day.weekday() >= 5)
        return week_end_days

    @staticmethod
    def working_delta(start_date, working_days):
        """
        Calculates the date based on a start date and the number of working days in the future

        :param start_date: Start date
        :param working_days: Number of woking days to the date we are interested
        :return: Date that is n working days from start date.
        """
        start_date = ACPHoliday.convert_to_date(start_date)
        working_days = int(working_days)
        first_guess = working_days + working_days/5*2 +4
        end_date = start_date + timedelta(days=first_guess)
        holidays = ACPHoliday.objects.filter(date__gte=start_date, date__lte=end_date)
        holiday_list = list()
        for holiday in holidays:
            holiday_list.append(holiday.date)
        not_completed = True
        end_date = start_date
        count = 0
        while not_completed:
            if end_date.weekday() < 5 and end_date not in holiday_list:
                count += 1
            if working_days == count:
                break
            end_date = end_date + timedelta(days=1)
        return end_date

    @staticmethod
    def get_working_days_for_month(year, month):
        """
        Calculate the amount of working days in a month

        :param year: Year
        :param month: month
        :return: Number of working days
        """
        try:
            last_day_of_month = monthrange(year, month)[1]
            start_date = date(year, month, 1)
            end_date = date(year, month, last_day_of_month)

            return ACPHoliday.get_working_days(start_date, end_date)
        except IllegalMonthError as e:
            raise ACPCalendarException(str(e))
